import hashlib
import datetime


def wrap_string(input_str, max_column_length=100, padding=0):
    words = input_str.split()
    lines = []
    current_line = words[0]

    for word in words[1:]:
        if len(current_line) + 1 + len(word) <= max_column_length:
            current_line += ' ' + word
        else:
            lines.append(current_line)
            current_line = word

    lines.append(current_line)
    join_char = "\n" + padding * " "
    formatted_output = join_char.join(lines)

    return formatted_output

# Dataclass, post init
# constructor -> .fromDict

class UniqueBlocksResult(object):

    statistics: dict
    unique_blocks: dict
    yara_rule: list

    def __init__(self) -> None:
        pass

    def generateYaraRule(self, min_ins=None, max_ins=None, min_bytes=None, max_bytes=None, required_per_sample=10, condition_required=7, wrap_at=0):
        cover = self.generateBlockCover(min_ins=min_ins, max_ins=max_ins, min_bytes=min_bytes, max_bytes=max_bytes, required_per_sample=required_per_sample)
        return self.renderRule(cover, condition_required, wrap_at=wrap_at)

    def renderRule(self, block_cover, condition_required, wrap_at=0):
        yara_blocks = block_cover["block_hashes"]
        block_hash_string = ",".join([pic_hash for pic_hash in yara_blocks])
        rule_identifier = hashlib.sha256(block_hash_string.encode()).hexdigest()[:16]
        yara_rule = f"rule mcrit_{rule_identifier} {{\n"
        yara_rule += "    meta:\n"
        yara_rule += "        author = \"MCRIT YARA Generator\"\n"
        yara_rule += f"        description = \"Code-based YARA rule composed from potentially unique basic blocks for the selected set of samples/family.\"\n"
        rule_date = datetime.datetime.utcnow().strftime("%Y-%m-%d")
        yara_rule += f"        date = \"{rule_date}\"\n"
        yara_rule += "    strings:\n"
        yara_rule += f"        // Rule generation selected {len(yara_blocks)} picblocks, covering {block_cover['num_samples_covered']}/{self.statistics['num_samples']} input sample(s).\n"
        for pichash, result in self.unique_blocks.items():
            if pichash not in yara_blocks:
                continue
            yarafied = f"        /* picblockhash: {pichash} - coverage: {len(result['samples'])}/{block_cover['num_samples_covered']} samples.\n"
            maxlen_ins = max([len(ins[1]) for ins in result["instructions"]])
            for ins in result["instructions"]:
                yarafied += f"         * {ins[1]:{maxlen_ins}} | {ins[2]} {ins[3]}\n"
            yarafied += "         */\n"
            if wrap_string:
                yarafied += f"        $blockhash_{pichash} = {{\n"
                yarafied +=  "            " + wrap_string(result["escaped_sequence"], max_column_length=80, padding=12) + "\n"
                yarafied += "        }\n"
            else:
                yarafied += f"        $blockhash_{pichash} = {{ " + result["escaped_sequence"] + " }\n"
            yara_rule += yarafied + "\n"
        yara_rule += "    condition:\n"
        yara_rule += f"        {min(len(yara_blocks), condition_required)} of them\n"
        yara_rule += "}\n"
        return yara_rule

    def generateBlockCover(self, min_ins=None, max_ins=None, min_bytes=None, max_bytes=None, required_per_sample=10):
        block_cover = {
            "block_hashes": [],
            "num_samples_covered": 0,
            "has_rule": False,
            "is_complete_cover": False
        }
        # we need to filter first, according to the desired parameters
        filtered_blocks = {}
        for block_hash, entry in self.unique_blocks.items():
            if min_ins and entry["length"] < min_ins:
                continue
            if max_ins and entry["length"] > max_ins:
                continue
            bytes_length = len(entry["escaped_sequence"].replace(" ", "")) // 2
            if min_bytes and bytes_length < min_bytes:
                continue
            if max_bytes and bytes_length > max_bytes:
                continue
            filtered_blocks[block_hash] = entry
        yara_rule_blocks = []
        sample_ids = [int(v) for v in self.statistics["by_sample_id"].keys()]
        sample_coverage = {sample_id: 0 for sample_id in sample_ids}
        samples_covered = set()
        while True:
            # calculate block_scores as how much benefit they bring, i.e. how many uncovered samples they can cover at once
            block_candidates = []
            for block_hash, entry in filtered_blocks.items():
                sample_ids_coverable = set(entry["samples"]).difference(samples_covered)
                if sample_ids_coverable and block_hash not in yara_rule_blocks:
                    candidate = {
                        "block_hash": block_hash,
                        "coverable": sample_ids_coverable,
                        "value": len(sample_ids_coverable),
                        "score": entry["score"]
                    }
                    block_candidates.append(candidate)
            # check if we are done yet, successful or not
            if len(samples_covered) == len(sample_ids):
                block_cover["has_rule"] = True
                block_cover["is_complete_cover"] = True
                break
            if len(block_candidates) == 0:
                if len(yara_rule_blocks) > 0:
                    block_cover["has_rule"] = True
                break
            # if not, choose the best block
            block_candidates.sort(key=lambda i: (i["value"], i["score"]))
            selected_block = block_candidates.pop()
            yara_rule_blocks.append(selected_block["block_hash"])
            # and update counters
            for sample_id in selected_block["coverable"]:
                sample_coverage[sample_id] += 1
            samples_covered = set([sample_id for sample_id, count in sample_coverage.items() if count >= required_per_sample])
            block_cover["num_samples_covered"] = len(samples_covered)
        block_cover["block_hashes"] = yara_rule_blocks
        return block_cover

    def toDict(self):
        # rebuild the original UniqueBlocks result here
        blocks_dict = {
            "statistics": self.statistics,
            "unique_blocks": self.unique_blocks,
            "yara_rule": self.yara_rule,
        }
        return blocks_dict

    @classmethod
    def fromDict(cls, entry_dict):
        blocks_result = cls()
        blocks_result.statistics = entry_dict["statistics"]
        blocks_result.unique_blocks = entry_dict["unique_blocks"]
        blocks_result.yara_rule = entry_dict["yara_rule"]
        return blocks_result

    def __str__(self):
        if self.statistics is not None:
            return "UniqueBlocksResult: {} Samples with {} unique blocks.".format(
                len(self.statistics["by_sample_id"]),
                len(self.unique_blocks)
            )
        else:
            return "UniqueBlocksResult: nothing parsed."



#!/usr/bin/env python3

import os
import uuid
import json
import logging
import hashlib
from random import sample
from collections import defaultdict
from itertools import zip_longest
from multiprocessing import Pool, cpu_count
from typing import Dict, List, Optional, TYPE_CHECKING, Tuple

import tqdm
from smda.common.BinaryInfo import BinaryInfo
from smda.common.SmdaFunction import SmdaFunction
from smda.common.SmdaReport import SmdaReport
from smda.Disassembler import Disassembler
from smda.SmdaConfig import SmdaConfig
from smda.common.SmdaInstruction import SmdaInstruction
from smda.intel.IntelInstructionEscaper import IntelInstructionEscaper

from mcrit.config.McritConfig import McritConfig
from mcrit.config.MinHashConfig import MinHashConfig
from mcrit.config.ShinglerConfig import ShinglerConfig
from mcrit.config.QueueConfig import QueueConfig
from mcrit.config.McritConfig import McritConfig
from mcrit.config.StorageConfig import StorageConfig
from mcrit.matchers.MatcherCross import MatcherCross
from mcrit.matchers.MatcherQuery import MatcherQuery
from mcrit.matchers.MatcherSample import MatcherSample
from mcrit.matchers.MatcherVs import MatcherVs
from mcrit.minhash.MinHasher import MinHasher
from mcrit.queue.QueueFactory import QueueFactory
from mcrit.queue.QueueRemoteCalls import NoProgressReporter, QueueRemoteCallee, Remote
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.storage.StorageFactory import StorageFactory

if TYPE_CHECKING:
    from mcrit.storage.StorageInterface import StorageInterface

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


class Worker(QueueRemoteCallee):
    def __init__(self, queue=None, config=None, storage: Optional["StorageInterface"] = None, profiling=False):
        self._worker_id = f"Worker-{uuid.uuid4()}"
        LOGGER.info(f"Starting as worker: {self._worker_id}")
        if config is None:
            config = McritConfig()

        if not queue:
            queue = QueueFactory().getQueue(config, consumer_id=self._worker_id)

        if profiling:
            print("[!] Running as profiled application.")
            profiling_path = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__ )), "..", "profiler"))
            os.makedirs(profiling_path, exist_ok=True)
        else:
            profiling_path = None
        super().__init__(queue, profiling_path)

        self.config = config
        self._storage_config = config.STORAGE_CONFIG
        self._minhash_config = config.MINHASH_CONFIG
        self._shingler_config = config.SHINGLER_CONFIG
        self._queue_config = config.QUEUE_CONFIG
        self.minhasher = MinHasher(config.MINHASH_CONFIG, config.SHINGLER_CONFIG)
        if storage:
            self._storage = storage
        else:
            self._storage = StorageFactory.getStorage(config)

    def  __enter__(self):
        return self

    def  __exit__(self, *args):
        # TODO unregister our worker_id from all in-progress jobs found in the queue
        self.queue.unregisterWorker()
        self.queue.release_all_jobs()

    #### STORAGE IO ####
    def getStorage(self):
        """Get an interface to the storage"""
        return self._storage

    def getStorageData(self):
        """Warning: This is intended for local debugging runs - storage may become huge"""
        results = {
            "config": {
                "minhash_config": self._minhash_config.toDict(),
                "shingler_config": self._shingler_config.toDict(),
                "storage_config": self._storage_config.toDict(),
                "queue_config": self._queue_config.toDict(),
            },
            "stats": self._storage.getStats(),
            "storage": self._storage.getContent(),
        }
        return results

    def setStorageData(self, storage_data):
        self._minhash_config = MinHashConfig.fromDict(storage_data["config"]["minhash_config"])
        self._shingler_config = ShinglerConfig.fromDict(storage_data["config"]["shingler_config"])
        self._storage_config = StorageConfig.fromDict(storage_data["config"]["storage_config"])
        self._queue_config = QueueConfig.fromDict(storage_data["config"]["queue_config"])
        mcrit_config = McritConfig()
        mcrit_config.MINHASH_CONFIG = self._minhash_config
        mcrit_config.SHINGLER_CONFIG = self._shingler_config
        mcrit_config.STORAGE_CONFIG = self._storage_config
        mcrit_config.QUEUE_CONFIG = self._queue_config
        # reinitialize
        self.minhasher = MinHasher(self._minhash_config, self._shingler_config)
        self._storage = StorageFactory.getStorage(mcrit_config)
        self._storage.setContent(storage_data["storage"])

    #### REDIRECTED FROM INDEX: MAIN WORKER FUNKTIONALITY ###

    @Remote(progress=True)
    def modifyFamily(self, family_id, update_information, progress_reporter=NoProgressReporter()):
        return self._storage.modifyFamily(family_id, update_information)

    @Remote(progress=True)
    def deleteFamily(self, family_id, keep_samples=False, progress_reporter=NoProgressReporter()):
        return self._storage.deleteFamily(family_id, keep_samples=keep_samples)

    @Remote(progress=True)
    def modifySample(self, sample_id, update_information, progress_reporter=NoProgressReporter()):
        return self._storage.modifySample(sample_id, update_information)

    @Remote(progress=True)
    def deleteSample(self, sample_id, progress_reporter=NoProgressReporter()):
        return self._storage.deleteSample(sample_id)

    def _addReport(self, smda_report, calculate_hashes=True, calculate_matches=False) -> "SampleEntry":
        sample_entry = self._storage.getSampleBySha256(smda_report.sha256)
        if sample_entry:
            LOGGER.info("Sample is already present in database: %s", sample_entry)
            return sample_entry
        sample_entry = self._storage.addSmdaReport(smda_report)
        if not sample_entry:
            return None
        LOGGER.info("Added %s", sample_entry)
        function_entries = self._storage.getFunctionsBySampleId(sample_entry.sample_id)
        LOGGER.info("Added %d function entries.", len(function_entries))
        job_id = None
        if calculate_hashes:
            self.updateMinHashesForSample(sample_entry.sample_id)
        return sample_entry

    # Reports PROGRESS
    @Remote(progress=True, file_locations=[0])
    def addBinarySample(self, binary, filename, family, version, is_dump, base_address, bitness, progress_reporter=NoProgressReporter()):
        binary_sha256 = hashlib.sha256(binary).hexdigest()
        sample_entry = self._storage.getSampleBySha256(binary_sha256)
        if sample_entry:
            LOGGER.info("Sample is already known with ID: %d", sample_entry.sample_id)
            return {"sample_info": sample_entry.toDict()}
        config = SmdaConfig()
        SMDA_REPORT = None
        DISASSEMBLER = Disassembler(config)
        LOGGER.info("Disassembling...")
        if is_dump:
            SMDA_REPORT = DISASSEMBLER.disassembleBuffer(binary, base_addr=base_address, bitness=bitness)
        else:
            SMDA_REPORT = DISASSEMBLER.disassembleUnmappedBuffer(binary)
        if filename is not None:
            SMDA_REPORT.filename = filename
        if family is not None:
            SMDA_REPORT.family = family
        if version is not None:
            SMDA_REPORT.version = version
        sample_entry = self._addReport(SMDA_REPORT)
        LOGGER.info("Disassembled and indexed sample: %s", sample_entry)
        if sample_entry is not None:
            return {"sample_info": sample_entry.toDict()}
        else:
            return None

    # Reports PROGRESS
    @Remote(progress=True)
    def rebuildIndex(self, progress_reporter=NoProgressReporter()):
        return self._storage.rebuildMinhashBandIndex(progress_reporter=progress_reporter)
    
    # Reports PROGRESS
    @Remote(progress=True)
    def recalculatePicHashes(self, progress_reporter=NoProgressReporter()):
        return self._storage.recalculateAllPicHashes(progress_reporter=progress_reporter)
    
     # Reports PROGRESS
    @Remote(progress=True)
    def recalculateMinHashes(self, progress_reporter=NoProgressReporter()):
        self._storage.deleteAllMinHashes(progress_reporter=progress_reporter)
        return self.updateMinHashes(None, progress_reporter=progress_reporter)

    # Reports PROGRESS
    @Remote(progress=True)
    def updateMinHashes(self, function_ids, progress_reporter=NoProgressReporter()):
        """Find unhashed functions in storage and calculate their MinHashes, optionally filter by function_ids or get function_entries passed directly"""
        if function_ids is None:
            # calculate all missing MinHashes in batches.
            unhashed_function_ids = self._storage.getUnhashedFunctions(None, only_function_ids=True)
            # to up to 10.000 function per batch
            LOGGER.info("Updating MinHashes: %d function entries have no MinHash yet.", len(unhashed_function_ids))
            total_batches = len(unhashed_function_ids) // self.config.MINHASH_CONFIG.MINHASH_GENERATION_WORKPACK_SIZE + (1 if len(unhashed_function_ids) % self.config.MINHASH_CONFIG.MINHASH_GENERATION_WORKPACK_SIZE else 0)
            progress_reporter.set_total(total_batches)
            for sliced_ids in zip_longest(*[iter(unhashed_function_ids)]*self.config.MINHASH_CONFIG.MINHASH_GENERATION_WORKPACK_SIZE):
                sliced_ids = [fid for fid in sliced_ids if fid is not None]
                unhashed_functions = self._storage.getUnhashedFunctions(sliced_ids)
                minhashes = self.calculateMinHashes(unhashed_functions, progress_reporter=progress_reporter)
                if minhashes:
                    self._storage.addMinHashes(minhashes)
                    LOGGER.info("Updated minhashes for %d function entries.", len(minhashes))
                progress_reporter.step()
        else:
            LOGGER.info("Updating MinHashes: %d function entries considered.", len(function_ids))
            total_batches = len(function_ids) // self.config.MINHASH_CONFIG.MINHASH_GENERATION_WORKPACK_SIZE + (1 if len(function_ids) % self.config.MINHASH_CONFIG.MINHASH_GENERATION_WORKPACK_SIZE else 0)
            progress_reporter.set_total(total_batches)
            for sliced_ids in zip_longest(*[iter(function_ids)]*self.config.MINHASH_CONFIG.MINHASH_GENERATION_WORKPACK_SIZE):
                sliced_ids = [fid for fid in sliced_ids if fid is not None]
                unhashed_functions = self._storage.getUnhashedFunctions(sliced_ids)
                LOGGER.info("Updating MinHashes: %d function entries have no MinHash yet.", len(unhashed_functions))
                minhashes = self.calculateMinHashes(unhashed_functions, progress_reporter=progress_reporter)
                if minhashes:
                    self._storage.addMinHashes(minhashes)
                    LOGGER.info("Updated minhashes for %d function entries.", len(minhashes))
                progress_reporter.step()
        # TODO if we do deferred calculation for a batch of minhashes, we might have to clear them here or address this where else updateMinHashes is used
        return len(minhashes)

    # Reports PROGRESS
    @Remote(progress=True)
    def updateMinHashesForSample(self, sample_id, progress_reporter=NoProgressReporter()):
        """Find unhashed functions in storage and calculate their MinHashes, optionally filter by function_ids or get function_entries passed directly"""
        function_entries = self._storage.getFunctionsBySampleId(sample_id)
        if function_entries:
            update_result = self.updateMinHashes([fe.function_id for fe in function_entries], progress_reporter=progress_reporter)
        else:
            LOGGER.info("Sample %d did not have any functions, proceeding.", sample_id)
            return 0
        if self.config.STORAGE_CONFIG.STORAGE_DROP_DISASSEMBLY:
            self._storage.deleteXcfgForSampleId(sample_id)
        return update_result

    # Reports PROGRESS
    @Remote(progress=True)
    def getUniqueBlocks(self, sample_ids, family_id=None, progress_reporter=NoProgressReporter()):
        # TODO we could propagate this progress reporter into the storage function for more fine grained progress tracking
        progress_reporter.set_total(1)
        blocks_result_dict = self._storage.getUniqueBlocks(sample_ids, progress_reporter=progress_reporter)
        blocks_result_dict["statistics"]["has_yara_rule"] = False
        blocks_result_dict["statistics"]["yara_covers"] = 0
        blocks_result_dict["statistics"]["has_complete_yara_rule"] = False
        unique_blocks = blocks_result_dict["unique_blocks"]
        # greedily produce a multi set cover of picblockhashes for sample_ids, i.e. a YARA rule :)
        yara_rule = []
        sample_coverage = {sample_id: 0 for sample_id in sample_ids}
        covers_required = 10
        functions_used = set()
        samples_covered = set()
        while True:
            # calculate block_scores as how much benefit they bring, i.e. how many uncovered samples they can cover at once
            block_candidates = []
            for block_hash, entry in unique_blocks.items():
                sample_ids_coverable = set(entry["samples"]).difference(samples_covered)
                if sample_ids_coverable and block_hash not in yara_rule:
                    candidate = {
                        "block_hash": block_hash,
                        "coverable": sample_ids_coverable,
                        "value": len(sample_ids_coverable),
                        "score": entry["score"]
                    }
                    block_candidates.append(candidate)
            # check if we are done yet, successful or not
            if len(samples_covered) == len(sample_ids):
                blocks_result_dict["statistics"]["has_yara_rule"] = True
                blocks_result_dict["statistics"]["has_complete_yara_rule"] = True
                break
            if len(block_candidates) == 0:
                if len(yara_rule) > 0:
                    blocks_result_dict["statistics"]["has_yara_rule"] = True
                break
            # if not, choose the best block
            block_candidates.sort(key=lambda i: (i["value"], i["score"]))
            selected_block = block_candidates.pop()
            yara_rule.append(selected_block["block_hash"])
            # and update counters
            for sample_id in selected_block["coverable"]:
                sample_coverage[sample_id] += 1
            samples_covered = set([sample_id for sample_id, count in sample_coverage.items() if count >= covers_required])
        blocks_result_dict["statistics"]["num_samples_covered"] = len(samples_covered)
        blocks_result_dict["yara_rule"] = yara_rule
        # enrich with escaped sequences
        sample_addr_borders = {}
        for sample_id in sample_ids:
            sample_entry = self._storage.getSampleById(sample_id)
            sample_addr_borders[sample_id] = {"lower": sample_entry.base_addr, "upper": sample_entry.base_addr + sample_entry.binary_size}
        for block_hash, entry in unique_blocks.items():
            sample_id = entry["sample_id"]
            escaped_sequences = []
            for instruction in entry["instructions"]:
                smda_instruction = SmdaInstruction(instruction)
                escaped_sequences.append(smda_instruction.getEscapedBinary(IntelInstructionEscaper, escape_intraprocedural_jumps=True, lower_addr=sample_addr_borders[sample_id]["lower"], upper_addr=sample_addr_borders[sample_id]["upper"]))
            unique_blocks[block_hash]["escaped_sequence"] = " ".join(escaped_sequences)
        blocks_result_dict["unique_blocks"] = unique_blocks
        progress_reporter.step()
        return blocks_result_dict

    # Reports PROGRESS
    @Remote(progress=True, json_locations=[0])
    def getMatchesForSmdaReport(
        self, 
        report_json, 
        minhash_threshold=None, 
        pichash_size=None, 
        band_matches_required=None, 
        progress_reporter=NoProgressReporter()
    ):
        matcher = MatcherQuery(
            self, 
            minhash_threshold=minhash_threshold, 
            pichash_size=pichash_size, 
            band_matches_required=band_matches_required,
            progress_reporter=progress_reporter
        )
        smda_report = SmdaReport.fromDict(report_json)
        match_report = matcher.getMatchesForSmdaReport(smda_report)
        return match_report

    # Reports PROGRESS
    @Remote(progress=True, file_locations=[0])
    def getMatchesForMappedBinary(
        self, 
        binary, 
        base_address, 
        minhash_threshold=None, 
        pichash_size=None, 
        band_matches_required=None,
        progress_reporter=NoProgressReporter()
    ):
        config = SmdaConfig()
        SMDA_REPORT = None
        DISASSEMBLER = Disassembler(config)
        SMDA_REPORT = DISASSEMBLER.disassembleBuffer(binary, base_address)
        matcher = MatcherQuery(
            self, 
            minhash_threshold=minhash_threshold, 
            pichash_size=pichash_size, 
            band_matches_required=band_matches_required,
            progress_reporter=progress_reporter
        )
        match_report = matcher.getMatchesForSmdaReport(SMDA_REPORT)
        return match_report

    # Reports PROGRESS
    @Remote(progress=True, file_locations=[0])
    def getMatchesForUnmappedBinary(
        self, 
        binary, 
        minhash_threshold=None, 
        pichash_size=None, 
        band_matches_required=None,
        progress_reporter=NoProgressReporter()
    ):
        config = SmdaConfig()
        SMDA_REPORT = None
        DISASSEMBLER = Disassembler(config)
        SMDA_REPORT = DISASSEMBLER.disassembleUnmappedBuffer(binary)
        matcher = MatcherQuery(
            self, 
            minhash_threshold=minhash_threshold, 
            pichash_size=pichash_size, 
            band_matches_required=band_matches_required,
            progress_reporter=progress_reporter
        )
        match_report = matcher.getMatchesForSmdaReport(SMDA_REPORT)
        return match_report

    # Reports PROGRESS
    @Remote(progress=True)
    def getMatchesForSample(
        self, 
        sample_id, 
        minhash_threshold=None, 
        pichash_size=None, 
        band_matches_required=None,
        progress_reporter=NoProgressReporter()
    ):
        matcher = MatcherSample(
            self, 
            minhash_threshold=minhash_threshold, 
            pichash_size=pichash_size, 
            band_matches_required=band_matches_required,
            progress_reporter=progress_reporter
        )
        match_report = matcher.getMatchesForSample(sample_id)
        return match_report

    # Reports PROGRESS
    @Remote(progress=True)
    def getMatchesForSampleVs(
        self,
        sample_id,
        other_sample_id,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        progress_reporter=NoProgressReporter(),
    ):
        matcher = MatcherVs(
            self, 
            minhash_threshold=minhash_threshold, 
            pichash_size=pichash_size, 
            band_matches_required=band_matches_required,
            progress_reporter=progress_reporter
        )
        match_report = matcher.getMatchesForSample(sample_id, other_sample_id)
        return match_report


    @Remote()
    def combineMatchesToCross(self, sample_to_job_id):
        child_results = []
        for job_id in sample_to_job_id.values():
            result = self.getResultForJob(job_id)
            if result is None:
                raise ValueError("Cannot evaluate Cross Compare. A child job failed.")
            # here we strip the results to sample level only, as function level matches are 
            # not required and may easily blow up memory
            result["matches"].pop("functions")
            child_results.append(result)
        return MatcherCross().create_result(child_results)


    def _groupItems(self, items, packsize=500):
        packed_items = []
        for lower_index in range(0, len(items) + packsize, packsize):
            sliced = items[lower_index : lower_index + packsize]
            if sliced:
                packed_items.append(sliced)
        return packed_items

    #### used by Worker(updateMinHashes) and MatcherQuery #####

    # Reports PROGRESS
    def calculateMinHashes(self, function_entries, progress_reporter=NoProgressReporter()):
        minhashes = []
        smda_functions = []
        LOGGER.info("Calculating MinHashes: hashing for %d function entries requested.", len(function_entries))
        for func in function_entries:
            binary_info = BinaryInfo(b"")
            binary_info.architecture = func.architecture
            smda_functions.append((func.function_id, SmdaFunction.fromDict(func.xcfg, binary_info=binary_info)))
        # filter down to functions that fulfill size requirements
        smda_functions = [
            (function_id, smda_function)
            for function_id, smda_function in smda_functions
            if self.minhasher.isMinHashableFunction(smda_function)
        ]
        LOGGER.info("Calculating MinHashes: %d function entries are indexable.", len(smda_functions))
        is_stepping = False
        if smda_functions:
            if self._minhash_config.MINHASH_POOL_INDEXING:
                packed_smda_functions = self._groupItems(smda_functions)
                if not progress_reporter.has_total_set():
                    progress_reporter.set_total(len(packed_smda_functions))
                    is_stepping = True
                with Pool(cpu_count()) as pool:
                    for result in tqdm.tqdm(
                        pool.imap_unordered(self.minhasher.calculateMinHashesFromStorage, packed_smda_functions),
                        total=len(packed_smda_functions),
                    ):
                        minhashes.extend(result)
                        if is_stepping:
                            progress_reporter.step()
            else:
                if not progress_reporter.has_total_set():
                    progress_reporter.set_total(len(smda_functions))
                    is_stepping = True
                for smda_function in tqdm.tqdm(smda_functions, total=len(smda_functions)):
                    minhashes.append(self.minhasher.calculateMinHashFromStorage(smda_function))
                    if is_stepping:
                        progress_reporter.step()
            LOGGER.info("Calculated minhashes for %d function entries!", len(minhashes))
        return minhashes

    ##### CONFIG CHANGES ####
    def updateMinHashThreshold(self, threshold):
        self.config.MINHASH_CONFIG.MINHASH_MATCHING_THRESHOLD = threshold
        self._minhash_config.MINHASH_MATCHING_THRESHOLD = threshold
        self.minhasher._minhash_config.MINHASH_MATCHING_THRESHOLD = threshold

    def updatePicHashSize(self, size):
        self.config.MINHASH_CONFIG.PICHASH_SIZE = size
        self._minhash_config.PICHASH_SIZE = size
        self.minhasher._minhash_config.PICHASH_SIZE = size

    def updateMinHasherConfig(self, config):
        self._storage_config = config.STORAGE_CONFIG
        self._minhash_config = config.MINHASH_CONFIG
        self._shingler_config = config.SHINGLER_CONFIG
        self.minhasher = MinHasher(config.MINHASH_CONFIG, config.SHINGLER_CONFIG)



import re
import logging
import datetime

import falcon

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex
from mcrit.server.utils import db_log_msg
from mcrit.queue.LocalQueue import Job

# TODO these should also return status and data in their json response

class JobResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get_collection(self, req, resp):
        # parse optional request parameters
        ascending = False
        if "ascending" in req.params:
            ascending = req.params["ascending"].lower().strip() == "true"
        method_filter = None
        if "method" in req.params:
            method_filter = req.params["method"]
        state_filter = None
        if "state" in req.params:
            state_filter = req.params["state"]
        query_filter = None
        if "filter" in req.params:
            query_filter = req.params["filter"]
        start_job_id = 0 
        if "start" in req.params:
            try:
                start_job_id = int(req.params["start"])
            except:
                pass
        limit_job_count = 0 
        if "limit" in req.params:
            try:
                limit_job_count = int(req.params["limit"])
            except:
                pass
        queue_data = self.index.getQueueData(start_index=start_job_id, limit=limit_job_count, method=method_filter, state=state_filter, filter=query_filter, ascending=ascending)
        resp.data = jsonify({"status": "successful", "data": queue_data})
        db_log_msg(self.index, req, f"JobResource.on_get_collection - success.")

    @timing
    def on_get_stats(self, req, resp):
        query_with_refresh = False
        if "with_refresh" in req.params:
            query_with_refresh = req.params["with_refresh"].lower().strip() == "true"
        queue_data = self.index.getQueueStats(refresh=query_with_refresh)
        resp.data = jsonify({"status": "successful", "data": queue_data})
        db_log_msg(self.index, req, f"JobResource.on_get_stats - success.")

    @timing
    def on_delete_collection(self, req, resp):
        # parse optional request parameters, to be used as an "AND" query
        method_filter = None
        if "method" in req.params:
            method_filter = req.params["method"]
        created_before = None
        if "created_before" in req.params:
            try:
                if len(req.params["created_before"]) == 10:
                    created_before = datetime.datetime.strptime(req.params["created_before"], "%Y-%m-%d")
                else:
                    created_before = datetime.datetime.strptime(req.params["created_before"], "%Y-%m-%dT%H:%M:%S")
            except:
                pass
        finished_before = None
        if "finished_before" in req.params:
            try:
                if len(req.params["finished_before"]) == 10:
                    finished_before = datetime.datetime.strptime(req.params["finished_before"], "%Y-%m-%d")
                else:
                    finished_before = datetime.datetime.strptime(req.params["finished_before"], "%Y-%m-%dT%H:%M:%S")
            except:
                pass
        # newest first
        result = self.index.deleteQueueData(method=method_filter, created_before=created_before, finished_before=finished_before)
        resp.data = jsonify({"status": "successful", "data": {"num_deleted": result}})
        db_log_msg(self.index, req, f"JobResource.on_delete_collection - success.")

    @timing
    def on_get(self, req, resp, job_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", job_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid JobIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_get - failed - invalid job_id.")
            return  
        data = self.index.getJobData(job_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get - success.")

    @timing
    def on_delete(self, req, resp, job_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", job_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid JobIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_delete - failed - invalid job_id.")
            return  
        result = self.index.deleteJob(job_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": {"num_deleted": result}})
        db_log_msg(self.index, req, f"JobResource.on_delete - success.")

    @timing
    def on_get_results(self, req, resp, result_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", result_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid ResultIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_get_results - failed - invalid result_id.")
            return 
        job_id = self.index.getJobIdForResult(result_id)
        job_data = self.index.getJobData(job_id)
        data = self.index.getResult(result_id)
        if "compact" in req.params and req.params["compact"].lower().strip() == "true":
            if job_data:
                job_info = Job(job_data, None)
                if job_info.is_matching_job or job_info.is_query_job:
                    data["matches"].pop("functions")
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get_results - success.")

    @timing
    def on_get_job_result(self, req, resp, job_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", job_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify()
            db_log_msg(self.index, req, f"JobResource.on_get_job_result - failed - invalid job_id.")
            return  
        job_data = self.index.getJobData(job_id)
        data = self.index.getResultForJob(job_id)
        if "compact" in req.params and req.params["compact"].lower().strip() == "true":
            if job_data:
                job_info = Job(job_data, None)
                if job_info.is_matching_job or job_info.is_query_job:
                    data["matches"].pop("functions")
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get_job_result - success.")

    @timing
    def on_get_result_job(self, req, resp, result_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", result_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid ResultIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_get_job_result - failed - invalid result_id.")
            return  
        job_id = self.index.getJobIdForResult(result_id)
        data = self.index.getJobData(job_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get_result_job - success.")



import re
import json
import logging
import zipfile

import falcon

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex
from mcrit.server.utils import db_log_msg

class StatusResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get(self, req, resp):
        resp.data = jsonify({"status": "successful", "data": {"message": "Welcome to MCRIT"}})
        db_log_msg(self.index, req, f"StatusResource.on_get - success.")

    @timing
    def on_get_status(self, req, resp):
        with_pichash = True if "with_pichash" in req.params and req.params["with_pichash"].lower() == "true" else False
        resp.data = jsonify({"status": "successful", "data": self.index.getStatus(with_pichash=with_pichash)})
        db_log_msg(self.index, req, f"StatusResource.on_get_status - success.")

    @timing
    def on_get_version(self, req, resp):
        resp.data = jsonify({"status": "successful", "data": self.index.getVersion()})
        db_log_msg(self.index, req, f"StatusResource.on_get_version - success.")

    @timing
    def on_get_config(self, req, resp):
        resp.status = falcon.HTTP_NOT_IMPLEMENTED
        db_log_msg(self.index, req, f"StatusResource.on_get_config - success / not implemented.")
        return

    @timing
    def on_get_export(self, req, resp):
        compress_data = True if "compress" in req.params and req.params["compress"].lower() == "true" else False
        try:
            exported_data = self.index.getExportData(compress_data=compress_data)
            resp.data = jsonify({"status": "successful", "data": exported_data})
            db_log_msg(self.index, req, f"StatusResource.on_get_export - success.")
        except MemoryError:
            resp.data = jsonify({"status": "failed", "data": {"message": "Export exceeded assigned maximum memory limit and was cancelled."}})
            db_log_msg(self.index, req, f"StatusResource.on_get_export - failed.")

    @timing
    def on_get_export_selection(self, req, resp, comma_separated_sample_ids=None):
        # NOTE if we encounter extreme cases (super long URLs), we might have to switch to post here.
        compress_data = True if "compress" in req.params and req.params["compress"].lower() == "true" else False
        exported_data = {}
        if re.match("^\d+(?:[\s]*,[\s]*\d+)*$", comma_separated_sample_ids):
            target_sample_ids = [int(sample_id) for sample_id in comma_separated_sample_ids.split(",")]
            exported_data = self.index.getExportData(target_sample_ids, compress_data=compress_data)
        resp.data = jsonify({"status": "successful", "data": exported_data})
        db_log_msg(self.index, req, f"StatusResource.on_get_export_selection - success.")

    @timing
    def on_post_import(self, req, resp):
        if not req.content_length:
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "POST request without body can't be processed."},
                }
            )
            resp.status = falcon.HTTP_400
            db_log_msg(self.index, req, f"StatusResource.on_post_import - failed - no POST body.")
            return
        import_data = json.loads(req.stream.read())
        import_report = self.index.addImportData(import_data)
        resp.data = jsonify({"status": "successful", "data": import_report})
        db_log_msg(self.index, req, f"StatusResource.on_post_import - success.")
        return

    @timing
    def on_post_respawn(self, req, resp):
        # this one has implicit "recalculation" because it nukes the whole DB
        self.index.respawn()
        resp.data = jsonify({"status": "successful", "data": {"message": "Successfully performed reset of MCRIT instance."}})
        db_log_msg(self.index, req, f"StatusResource.on_post_respawn - success.")

    @timing
    def on_get_complete_minhashes(self, req, resp):
        minhash_report = self.index.updateMinHashes(None, force_recalculation=True)
        resp.data = jsonify({"status": "successful", "data": minhash_report})
        db_log_msg(self.index, req, f"StatusResource.on_get_complete_minhashes - success.")
        return

    @timing
    def on_get_rebuild_index(self, req, resp):
        index_report = self.index.rebuildIndex(force_recalculation=True)
        resp.data = jsonify({"status": "successful", "data": index_report})
        db_log_msg(self.index, req, f"StatusResource.on_get_rebuild_index - success.")
        return

    @timing
    def on_get_recalculate_pichashes(self, req, resp):
        index_report = self.index.recalculatePicHashes(force_recalculation=True)
        resp.data = jsonify({"status": "successful", "data": index_report})
        db_log_msg(self.index, req, f"StatusResource.on_get_recalculate_pichashes - success.")
        return

    @timing
    def on_get_recalculate_minhashes(self, req, resp):
        index_report = self.index.recalculateMinHashes(force_recalculation=True)
        resp.data = jsonify({"status": "successful", "data": index_report})
        db_log_msg(self.index, req, f"StatusResource.on_get_recalculate_minhashes - success.")
        return

    @staticmethod
    def _get_search_args(params):
        result = {
            "search_term": params["query"],
            "cursor": params.get("cursor", None),
            "sort_by": params.get("sort_by", None),
            "is_ascending": params.get("is_ascending", "true").lower() != "false",
        }
        try: 
            result["limit"] = int(params.get("limit"))
        except:
            pass
        return result

    @timing
    def on_get_search_families(self, req, resp):
        args = self._get_search_args(req.params)
        resp.data = jsonify({"status": "successful", "data": self.index.getFamilySearchResults(**args)})
        db_log_msg(self.index, req, f"StatusResource.on_get_search_families - success.")

    @timing
    def on_get_search_samples(self, req, resp):
        args = self._get_search_args(req.params)
        resp.data = jsonify({"status": "successful", "data": self.index.getSampleSearchResults(**args)})
        db_log_msg(self.index, req, f"StatusResource.on_get_search_samples - success.")

    @timing
    def on_get_search_functions(self, req, resp):
        args = self._get_search_args(req.params)
        resp.data = jsonify({"status": "successful", "data": self.index.getFunctionSearchResults(**args)})
        db_log_msg(self.index, req, f"StatusResource.on_get_search_functions - success.")



"""Search query parser

version 2022-08-08

This search query parser uses the excellent Pyparsing module
(http://pyparsing.sourceforge.net/) to parse search queries by users.
It handles:

* 'and', 'or' and implicit 'and' operators;
* parentheses;
* quoted strings;
* <field>:<value>
* <field>:<operator><value> for the following operators:
    <, <=, >, >=, =, !=, ? (search in a single field), !?


Requirements:
* Python
* Pyparsing


This work is based on 
https://github.com/pyparsing/pyparsing/blob/master/examples/searchparser.py
and modified by Manuel Blatt, 2022
The original license can be found below.

-------------------------------------------------------------------------------
Copyright (c) 2006, Estrate, the Netherlands
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
* Neither the name of Estrate nor the names of its contributors may be used
  to endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

CONTRIBUTORS:
- Steven Mooij
- Rudolph Froger
- Paul McGuire
-------------------------------------------------------------------------------

"""

from functools import lru_cache
from string import whitespace
from pyparsing import (
    Word,
    CharsNotIn,
    CharsNotIn,
    Group,
    ZeroOrMore,
    OneOrMore,
    QuotedString,
    oneOf,
    Suppress,
    identchars,
    identbodychars,
    Forward,
    Keyword,
    StringEnd,
)

from mcrit.index.SearchQueryTree import AndNode, NodeType, NotNode, OrNode, SearchConditionNode, SearchTermNode

class SearchQueryParser:
    def __init__(self):
        self._methods = {
            "and": self._build_tree_and,
            "or": self._build_tree_or,
            "not": self._build_tree_not,
            "parenthesis": self._build_tree_parenthesis,
            "one_filter": self._build_tree_one_filter
        }
        self._parser = self._get_parser()
    
    def _get_parser(self):
        search_term = QuotedString('"', esc_char="\\", esc_quote='\\"') \
            | QuotedString("'", esc_char="\\", esc_quote="\\'") \
            | ZeroOrMore(Suppress(whitespace))+CharsNotIn(whitespace+"()")
        identifier = Word(identchars, identbodychars+".")
        operator = oneOf("< <= > >= = != ? !?")
        condition_compare = identifier + (":" + operator + search_term).leave_whitespace()
        condition_equal = identifier + (":" + search_term).leave_whitespace()
        one_filter = Group(condition_compare | condition_equal | search_term).setResultsName("one_filter")

        operatorOr = Forward()

        operatorParenthesis = (
            Group(Suppress("(") + operatorOr + Suppress(")")).setResultsName(
                "parenthesis"
            )
            | one_filter
        )

        operatorNot = Forward()
        operatorNot << (
            Group(Suppress(Keyword("NOT")) + operatorNot).setResultsName(
                "not"
            )
            | operatorParenthesis
        )

        operatorAnd = Forward()
        operatorAnd << (
            Group(
                operatorNot + Suppress(Keyword("AND")) + operatorAnd
            ).setResultsName("and")
            | Group(
                operatorNot + OneOrMore(~oneOf("AND OR") + operatorAnd)
            ).setResultsName("and")
            | operatorNot
        )

        operatorOr << (
            Group(
                operatorAnd + Suppress(Keyword("OR")) + operatorOr
            ).setResultsName("or")
            | operatorAnd
        )

        parser = operatorOr + StringEnd()
        return parser
    
    def _build_tree_or(self, argument) -> NodeType:
        assert len(argument) == 2
        children = [self._build_tree(child) for child in argument]
        return OrNode(children)

    def _build_tree_and(self, argument) -> NodeType:
        assert len(argument) == 2
        children = [self._build_tree(child) for child in argument]
        return AndNode(children)

    def _build_tree_parenthesis(self, argument) -> NodeType:
        assert len(argument) == 1
        return self._build_tree(argument[0])

    def _build_tree_not(self, argument) -> NodeType:
        assert len(argument) == 1
        return NotNode(self._build_tree(argument[0]))

    def _build_tree_one_filter(self, argument) -> NodeType:
        if len(argument) == 1:
            return SearchTermNode(argument[0])
        if len(argument) == 3:
            return SearchConditionNode(argument[0], "", argument[-1])
        if len(argument) == 4:
            return SearchConditionNode(argument[0], argument[-2], argument[-1])
        raise ValueError()
    
    def _build_tree(self, argument) -> NodeType:
        return self._methods[argument.getName()](argument)

    @lru_cache(maxsize=100)
    def parse(self, string:str) -> NodeType:
        if string.strip(whitespace) == "":
            return AndNode([])
        raw_result = self._parser.parse_string(string)
        assert len(raw_result) == 1
        result = self._build_tree(raw_result[0])
        return result


from typing import Dict, List, Optional

from mcrit.storage.SampleEntry import SampleEntry


class FamilyEntry(object):

    family_id: int
    family_name: str
    num_samples: int
    num_functions: int
    num_library_samples: int
    # This is not supposed to be stored in storage
    samples: Optional[Dict[int, SampleEntry]]

    def __init__(self, family_name="", family_id=0, num_samples=0, num_functions=0, num_library_samples=0, samples=None):
        self.family_id = family_id
        self.family_name = family_name
        self.num_samples = num_samples
        self.num_functions = num_functions
        self.num_library_samples = num_library_samples 
        self.samples = samples

    @property
    def is_library(self):
        return self.num_library_samples > 0

    @property
    def family(self):
        return self.family_name

    def toDict(self):
        family_entry = {
            "family_id": self.family_id,
            "family_name": self.family_name,
            "num_samples": self.num_samples,
            "num_functions": self.num_functions,
            "num_library_samples": self.num_library_samples,
        }
        if self.samples is not None:
            family_entry["samples"] = {id: sample.toDict() for id, sample in self.samples.items()}
        return family_entry

    @classmethod
    def fromDict(cls, entry_dict: Dict):
        family_entry = cls(None) #type: ignore
        family_entry.family_id = entry_dict["family_id"]
        family_entry.family_name = entry_dict["family_name"]
        family_entry.num_samples = entry_dict["num_samples"]
        family_entry.num_functions = entry_dict["num_functions"]
        family_entry.num_library_samples = entry_dict["num_library_samples"]
        samples = entry_dict.get("samples", None)
        if samples is not None:
            family_entry.samples = {id: SampleEntry.fromDict(sample) for id, sample in samples.items()}
        return family_entry

    def __str__(self):
        return "Famliy {} ({}): ".format(
            self.family_id, self.family_name
        )




from dataclasses import asdict, dataclass, field

from mcrit.config.ConfigInterface import ConfigInterface

@dataclass
class GunicornConfig(ConfigInterface):
    # introduce a switch for gunicorn
    USE_GUNICORN: bool = False
    # specified as "<host>:<port>" to be able to pass it easily into gunicorn config
    BIND: str = "0.0.0.0:8000"
    # The number of workers gunicorn will spin up
    WORKERS: int = 4
    # The number of threads in each worker
    THREADS: int = 8
    # timeout before silent workers are killed
    TIMEOUT: int = 120